@@ -1,25 +1,25 @@
 package Log::Log4perl::Appender::Raven;
-$Log::Log4perl::Appender::Raven::VERSION = '0.004';
+$Log::Log4perl::Appender::Raven::VERSION = '0.005';
 use Moose;
 
 use Carp;
 use Data::Dumper;
+use Digest::MD5;
 use Sentry::Raven;
 use Log::Log4perl;
 use Devel::StackTrace;
+use Safe;
+use Text::Template;
 
+## Configuration
 has 'sentry_dsn' => ( is => 'ro', isa => 'Maybe[Str]' );
 has 'sentry_timeout' => ( is => 'ro' , isa => 'Int' ,required => 1 , default => 1 );
+has 'sentry_culprit_template' => ( is => 'ro', isa => 'Str', required => 1 , default => '{$function}');
 has 'infect_die' => ( is => 'ro' , isa => 'Bool', default => 0 );
-
-has 'raven' => ( is => 'ro', isa => 'Sentry::Raven', lazy_build => 1);
-
 # STATIC CONTEXT
 has 'context' => ( is => 'ro' , isa => 'HashRef', default => sub{ {}; });
-
 # STATIC TAGS. They will go in the global context.
 has 'tags' => ( is => 'ro' ,isa => 'HashRef', default => sub{ {}; });
-
 # Log4Perl MDC key to look for tags
 has 'mdc_tags' => ( is => 'ro' , isa => 'Maybe[Str]' , default => 'sentry_tags' );
 # Log4perl MDC key to look for extra
@@ -29,6 +29,14 @@ has 'mdc_user'  => ( is => 'ro' ,isa => 'Maybe[Str]' , default => 'sentry_user'
 # Log4perl MDC key to look for http data.
 has 'mdc_http' => ( is => 'ro' , isa => 'Maybe[Str]' , default => 'sentry_http' );
 
+## End of configuration
+
+# Operation objects
+has 'raven' => ( is => 'ro', isa => 'Sentry::Raven', lazy_build => 1);
+has 'culprit_text_template' => ( is => 'ro', isa => 'Text::Template' , lazy_build => 1);
+has 'safe' => ( is => 'ro' , isa => 'Safe', lazy_build => 1);
+
+
 my %L4P2SENTRY = ('ALL' => 'info',
                   'TRACE' => 'debug',
                   'DEBUG' => 'debug',
@@ -101,6 +109,28 @@ See perldoc Log::Log4perl::Appender::Raven, section 'CODE WIHTOUT LOG4PERL'
     }
 }
 
+sub _build_safe{
+    # We do not authorize anything.
+    return Safe->new();
+}
+
+{
+    # The fallback culprint template will signal itself as such in sentry.
+    my $FALLBACK_CULPRIT_TEMPLATE = 'FALLBACK {$function}';
+    sub _build_culprit_text_template{
+        my ($self) = @_;
+        my $tmpl = Text::Template->new( TYPE => 'STRING',
+                                        SOURCE => $self->sentry_culprit_template(),
+                                    );
+        unless( $tmpl->compile() ){
+            warn "Cannot compile template from '".$self->sentry_culprit_template()."' ERROR:".$Text::Template::ERROR.
+                " - Will fallback to hardcoded '".$FALLBACK_CULPRIT_TEMPLATE."'";
+            $tmpl = Text::Template->new( TYPE => 'STRING', SOURCE => $FALLBACK_CULPRIT_TEMPLATE);
+            $tmpl->compile() or die "Invalid fallback template ".$FALLBACK_CULPRIT_TEMPLATE;
+        }
+        return $tmpl;
+    }
+}
 
 sub _build_raven{
     my ($self) = @_;
@@ -139,7 +169,7 @@ sub log{
     # We are 4 levels down after the standard Log4perl caller_depth
     my $caller_offset = Log::Log4perl::caller_depth_offset( $Log::Log4perl::caller_depth + 4 );
 
-    ## Stringify arguments NOW. This avoid sending huuge objects when
+    ## Stringify arguments NOW (no_refs => 1). This avoids sending huuuuuge objects when
     ## Serializing this stack trace inside Sentry::Raven
     my $caller_frames = Devel::StackTrace->new( no_refs => 1);
     {
@@ -149,11 +179,18 @@ sub log{
         $caller_frames->frames(@frames);
     }
 
-    my $sentry_culprit = 'main';
+    my $call_depth = $caller_offset;
+
+    # Defaults caller properties
+    my $caller_properties = {
+        function => 'main',
+    };
+    my @log_call_info = caller($call_depth - 1);
+    $caller_properties->{line} = $log_call_info[2] || 'NOTOPLINE';
+
     {
-        my $call_depth = $caller_offset;
         # Go up the caller ladder until the first non eval
-        while( my @caller_info = caller($call_depth++) ){
+        while( my @caller_info = caller($call_depth) ){
 
           # Skip evals and __ANON__ methods.
           # The anon method will make that compatible with the new Log::Any (>0.15)
@@ -163,12 +200,17 @@ sub log{
                   || ( scalar(reverse($caller_string)) =~ /^__NONA__/ )
                   #  ^ This test for the caller string to end with __ANON__ , but faster.
                 ){
-            # This is good.
-            # Subroutine name, or filename, or just main
-            $sentry_culprit = $caller_info[3] || $caller_info[1] || 'main';
-            last;
+              # This is good.
+              # Subroutine name, or filename, or just main
+              $caller_properties->{function} = $caller_info[3] || $caller_info[1] || 'main';
+              # For other properties, we are interested in the place where $log->something
+              # was called, not were the caller of $log->something was called from
+              my @log_call_info = caller($call_depth - 1);
+              $caller_properties->{line} = $log_call_info[2] || 'NOLINE';
+              last;
           }
-        }
+          $call_depth++;
+      }
     }
 
     my $tags = {};
@@ -191,6 +233,21 @@ sub log{
         $http = Log::Log4perl::MDC->get($mdc_http);
     }
 
+    # Calculate the culprit from the template
+    my $sentry_culprit = $self->culprit_text_template->fill_in(
+        SAFE => $self->safe(),
+        HASH => {
+            %{$caller_properties},
+            message => $sentry_message,
+            sign => sub{
+                my ($string, $offset, $length) = @_;
+                defined( $string ) || ( $string = '' );
+                defined( $offset ) || ( $offset = 0 );
+                defined( $length ) || ( $length = 4 );
+                return substr(Digest::MD5::md5_hex(substr($string, $offset, $length)), 0, 4);
+            }
+        });
+
     # OK WE HAVE THE BASIC Sentry options.
     $self->raven->capture_message($sentry_message,
                                   logger => $sentry_logger,
@@ -279,6 +336,55 @@ Example:
   log4perl.appender.Raven.layout=${layout_class}
   log4perl.appender.Raven.layout.ConversionPattern=${layout_pattern}
 
+=head2 Configuring the culprit string
+
+By default, this appender will calculate the Sentry culprit to be
+the fully qualified name of the function that called the log method, as Sentry
+recommends.
+
+If you require more flexibility and precision in your culprit, you can
+configure it as a template. For instance:
+
+  log4perl.appender.Raven.sentry_culprit_template={$function}-{$line}
+
+The default is '{$function}', as Sentry prescribes. But most people will probably
+be more happy with the added {$line} element, as it makes discriminating between culprits easier.
+
+The template format follows L<Text::Template> and the available variables and functions are as follow:
+
+=over
+
+=item function
+
+The fully qualified name of the function that called the log method.
+
+=item line
+
+The line at which the log method was called
+
+=item message
+
+The Log4perl generated message. Keep in mind that this is the message AFTER it has been calculated by
+the layout pattern.
+
+=item sign($string, $offset, $length)
+
+A function that calculates a small (4 chars) signature of the given string. $string, $offset
+and $length are optional.
+
+This is useful for instance if some part of your code manage errors in a centralized way, or in other
+terms if the place at which you call '$log->error()' can output various messages.
+To help discriminating between culprit, you can for instance configure your culprit template:
+
+
+  log4perl.appender.Raven.sentry_culprit_template={$function}-{$line}-{sign($message, 30, 4)}
+
+Note that in the example, we look at a part of the message after the 30th character, which
+helps skipping the common message parts defined by your message layout. Adjust this number (30)
+to make sure you pick a substring of your message in a meaningful area.
+
+=back
+
 =head2 Timeout
 
 The default timeout is 1 second. Feel free to bump it up. If sending an event
